## The estimation of normal mixtures with latent variables
## by Gideon Magnus and Jan R. Magnus

#save.image()
#load(".Rdata")


if(!require(mvtnorm)) {install.packages("mvtnorm"); require(mvtnorm)}
if(!require(numDeriv)) {install.packages("numDeriv"); require(numDeriv)}
if(!require(matrixcalc)) {install.packages("matrixcalc"); require(matrixcalc)}
if(!require(MASS)) {install.packages("MASS"); require(MASS)} # needed for generalized inverse
if(!require(xtable)) {install.packages("xtable"); require(xtable)} # needed for generalized inverse
if(!require(WriteXLS)) {install.packages("WriteXLS"); require(WriteXLS)} # needed for generalized inverse

source("logsumfun.R")
source("logfracfun.R")
source("extractparam.R")
source("make.ident.params.R")
source("kernel1.R")
source("gradient1.R")
source("hessian1.R")
source("rnorm_ALB.R")




all_models=list()
nr_models=8
gammavec=c(.2,.15,.1,.05,.01,.01,.01,.01)



# n: nr of variables
# TT: nr of time periods
# M: nr of latent factors
# g: nr of regimes

TT=200
g=2


## START OF LOOP
for (modl in 6:nr_models)
{
model=list()

set.seed(modl)

if (modl==1)
{
  n=4
  M=2
}

if (modl==2)
{
  n=6
  M=3
}

if (modl==3)
{
  n=8
  M=4
  }

if (modl==4)
{
  n=10
  M=5
  }

if (modl==5)
{
  n=20
  M=5
}

if (modl==6)
{
  n=30
  M=5
}

if (modl==7)
{
  n=40
  M=5
}

if (modl==8)
{
  n=50
  M=5
}


model$M=M
model$n=n


source("fake_data_generation.R")


inclmat=is.na(XX)
storage.mode(inclmat)="integer"

param0=c(xi0[2:g],c(mu0),c(VVtildemat0),BBvec0,psitilde0)
parameters = param0
nrparam=length(parameters)

ind0=c(1:(g-1)) # xi
ind1=g-1+c(1:(g*M)) # mu
ind2=g-1+c((g*M+1):(g*M+g*M*(M+1)/2)) #VVtilde
ind3=g-1+c((g*M+g*M*(M+1)/2+1):(g*M+g*M*(M+1)/2 +M*n - .5*(M^2+M))) # B
ind4=g-1+c((g*M+g*M*(M+1)/2 + M*n - .5*(M^2+M) +1):(g*M+g*M*(M+1)/2 + M*n - .5*(M^2+M) +n)) # psitilde

# ind.select: exclude the fixed entries of BB from a "full" vector of parameters
temp = g-1+c((g*M+g*M*(M+1)/2+1):(g*M+g*M*(M+1)/2 + M*n)) 
mattemp = matrix(temp,n,M)
ind3temp=c()
for (j in 1:M) {ind3temp=c(ind3temp,mattemp[(j+1):n,j])}
ind4temp=g-1+c((g*M+g*M*(M+1)/2 + M*n+1):(g*M+g*M*(M+1)/2 + M*n+n)) # psitilde
ind.select=c(ind0,ind1,ind2,ind3temp,ind4temp)

model$param0=param0
model$XX=XX



listtemp =  extractparam(parameters,XX,kappa1,kappa2,M,g)

out = kernel1(parameters,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)

## testing that the gradient is correct:
#test1 = gradient1(parameters,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g,ind.select=ind.select)
#test2 = grad(kernel1,parameters,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)
#temp= test1-test2
#round(temp,2)

## testing that the hessian is correct:
#hesstest1 = hessian(kernel1,parameters,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g) # using NumDeriv
#hesstest2 = hessian1(parameters,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g,ind.select=ind.select) # using analytical
#test=hesstest1-hesstest2
#round(test,2)
 


## OPTIMIZATION
itime=proc.time()[3]
out = optim(par=param0,fn=kernel1,gr=gradient1,method="BFGS",control=list(trace=TRUE,maxit=5000,REPORT=1,reltol=1e-12),XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g,ind.select=ind.select)
ctime = proc.time()[3]
total.time.optim = (ctime-itime)/60
cat('  Total Time Elapsed: ',round((ctime-itime)/60,2),'\n')

model$optim_time = total.time.optim

param.max1 = out$par
kernel1(param.max1,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)
param.list.max1 = extractparam(param.max1,XX,kappa1,kappa2,M,g)

param.max2 = make.ident.params(param.max1,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)
kernel1(param.max2,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)
param.list.max2 = extractparam(param.max2,XX,kappa1,kappa2,M,g)


# plot(param.max1,type="l")
# lines(param.max2,col="red")
# lines(param0,col="blue")


param.max=param.max2

# test1 = gradient1(param.max2,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)
# round(test1,1)

param.list.max = extractparam(param.max,XX,kappa1,kappa2,M,g)

model$param.max=param.max


########################################################################################################
## calculate 2 matrices: # 1) hessian # 2) expected outer product of score = expected hessian

## hessian
H1 = hessian1(param.max,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g,ind.select=ind.select) 


## expected outer product of score
REP=20000
XXtemp = matrix(0,TT,n)
H2=matrix(0,nrparam,nrparam)

itime=proc.time()[3]
for (rep in 1:REP)
{
if (rep%%5==0) {cat(rep," / ",REP,"\n")}
  
  for (t in 1:TT)
  {
    ind = sample(c(1:g),1,prob=pi0)
    XXtemp[t,] = mm0[,ind] + tcholWW0[,,ind]%*%rnorm(n)
  }
  XXtemp[inclmat==1]=NA
  
  gradtemp = gradient1(param.max,XX=XXtemp,kappa1=kappa1,kappa2=kappa2,M=M,g=g,ind.select=ind.select)

  H2 = H2 + tcrossprod(gradtemp)/REP
}
ctime = proc.time()[3]
timetemp = (ctime-itime)/60

model$inf.mat.time = timetemp


H1.inv=ginv(H1)
H2.inv=ginv(H2)

# plot(sqrt(diag(H1.inv)),type="l")
# lines(sqrt(diag(H2.inv)),col="blue")


bla=FALSE
i=50
H1.ginv=ginv(H1)

while (bla==FALSE)
{
  var1 = H1.ginv + diag(nrparam)/(10^i)
  test   = try(chol(var1),silent=TRUE)
  if (is.matrix(test)) {bla=TRUE}
  i=i-1
}
cat(i)



bla=FALSE
i=50
H2.ginv=ginv(H2)

while (bla==FALSE)
{
  var2 = H2.ginv + diag(nrparam)/(10^i)
  test   = try(chol(var2),silent=TRUE)
  if (is.matrix(test)) {bla=TRUE}
  i=i-1
}
cat(i)


plot(sqrt(diag(var1)),type="l")
lines(sqrt(diag(var2)),col="blue")



model$H1=H1
model$H2=H2
model$var1=var1
model$var2=var2







##########################################################################################
## run a random walk MCMC sampler
vartemp = gammavec[modl]*var1
tcholvar = t(chol(vartemp))

REP=100000
paramdraw=matrix(0,REP,nrparam)
paramold=param.max
kernelold= -kernel1(paramold,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)
nraccept=0
do.ident=TRUE

itime=proc.time()[3]
for (rep in 1:REP)
{
  if (rep%%10==0) {cat(rep," / ",REP," accept rate: ",nraccept/rep,"\n")}
  
  paramcand  = paramold + tcholvar%*%rnorm(nrparam)
  kernelcand = -kernel1(paramcand,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)
  
  alp = min(1,exp(kernelcand-kernelold))
  
  accept = rbinom(1,1,alp)
  if (accept)
  {
    paramold=paramcand
    kernelold=kernelcand
    nraccept=nraccept+1
      }
  
  paramtemp=paramold
  if (do.ident) {paramtemp = make.ident.params(paramtemp,XX=XX,kappa1=kappa1,kappa2=kappa2,M=M,g=g)  }
  
  paramdraw[rep,]=paramtemp
}

ctime = proc.time()[3]
cat('  Total Time Elapsed: ',round((ctime-itime)/60,2),'\n')

###################################################

timetemp = (ctime-itime)/60
model$MCMC.time = timetemp

model$paramdraw=paramdraw
model$acceptrate = nraccept/REP

varMCMC= var(paramdraw)
maxtemp=max(sqrt(diag(varMCMC))); maxtemp=max(maxtemp,sqrt(diag(var1))); maxtemp=max(maxtemp,sqrt(diag(var2)))

nametemp=paste0("posterior_std_devs",modl,".pdf")
pdf(nametemp)
plot(sqrt(diag(varMCMC)),type="l",lwd=2,ylim=c(0,maxtemp))
lines(sqrt(diag(var1)),col="blue",lwd=2) ## exact hessian
lines(sqrt(diag(var2)),col="red",lwd=2)  ## score outer product
legend(nrparam*.6,maxtemp,c("true","-H","E(qq')"),col=c("black","blue","red"),pch=18)
grid()
dev.off()



mattemp=cbind(sqrt(diag(varMCMC)),sqrt(diag(var1)),sqrt(diag(var2)))
colnames(mattemp) = c("true","hessian","inf. matrix")
nametemp=paste0("posterior_std_devs",modl,".xls")
WriteXLS(as.data.frame(mattemp),ExcelFileName = nametemp)

logdiff1 = abs(log(sqrt(diag(varMCMC))) - log(sqrt(diag(var1))))
logdiff2 = abs(log(sqrt(diag(varMCMC))) - log(sqrt(diag(var2))))
round(100*mean(logdiff1),1) # hessian 
round(100*mean(logdiff2),1) # gradient

diff1 = sqrt(diag(varMCMC)) - sqrt(diag(var1))
diff2 = sqrt(diag(varMCMC)) - sqrt(diag(var2))

mattemp = matrix(0,2,3)
rownames(mattemp)=c("Hessian","Inf. matrix")
colnames(mattemp)=c("mean abs. diff.","sd","mean abs. log diff.")
mattemp[1,1] = mean(abs(diff1))
mattemp[1,2] = sd(diff1)
mattemp[1,3] = mean(logdiff1)
mattemp[2,1] = mean(abs(diff2))
mattemp[2,2] = sd(diff2)
mattemp[2,3] = mean(logdiff2)
nametemp=paste0("differences_sd",modl,".xls")
WriteXLS(as.data.frame(mattemp),ExcelFileName = nametemp,row.names=TRUE)

nametemp=paste0("differences_sd",modl,".tex")
test = xtable(mattemp,label="tab:differences_corr",align=c("lccc"),caption="posterior standard deviations",digits=4)
print(test,table.placement="H",floating=TRUE,type="latex",file=nametemp)

  

cor.elem.true = varMCMC[upper.tri(cov2cor(varMCMC),diag=FALSE)]
cor.elem.1  = var1[upper.tri(cov2cor(var1),diag=FALSE)]
cor.elem.2  = var2[upper.tri(cov2cor(var2),diag=FALSE)]

mattemp=cbind(cor.elem.true,cor.elem.1,cor.elem.2)
colnames(mattemp) = c("true","hessian","inf. matrix")
nametemp=paste0("posterior_correlations",modl,".xls")
WriteXLS(as.data.frame(mattemp),ExcelFileName = nametemp)

diff1 = cor.elem.1 - cor.elem.true
diff2 = cor.elem.2 - cor.elem.true

mattemp = matrix(0,2,2)
rownames(mattemp)=c("Hessian","Inf. matrix")
colnames(mattemp)=c("mean abs. diff.","sd")
mattemp[1,1] = mean(abs(diff1))
mattemp[1,2] = sd(diff1)
mattemp[2,1] = mean(abs(diff2))
mattemp[2,2] = sd(diff2)
nametemp=paste0("differences_corr",modl,".xls")
WriteXLS(as.data.frame(mattemp),ExcelFileName = nametemp,row.names=TRUE)

nametemp=paste0("differences_corr",modl,".tex")
test = xtable(mattemp,label="tab:differences_corr",align=c("lcc"),caption="posterior correlations",digits=4)
print(test,table.placement="H",floating=TRUE,type="latex",file=nametemp)



all_models[[modl]] = model
} ## for modl in 1:nr_models



################################################################################################
################################################################################################










## differences

i=12

plot(paramdraw[,i],type="l",main=i)
lines(rep(param.max[i],REP),col="magenta",lwd=4)
i=i+1
 
hist(paramdraw[,i],col="magenta",freq=FALSE,breaks=100)




# create table of run times

# optim
mattemp=matrix(0,nr_models,1)
for (i in 1:nr_models)
{
  mattemp[i,1] =  all_models[[i]]$optim_time
}

# information matrix
mattemp=matrix(0,nr_models,1)
for (i in 1:nr_models)
{
  mattemp[i,1] =  all_models[[i]]$inf.mat.time
}

# MCMC
mattemp=matrix(0,nr_models,1)
for (i in 1:nr_models)
{
  mattemp[i,1] =  all_models[[i]]$MCMC.time
}

